;**********************************************************************************
;File: LowLevel.asm
;Initialization of hardware and some low level routines
;**********************************************************************************
; if !(TX_KBPS == 800 || TX_KBPS == 400 || TX_KBPS == 250)
 if !(TX_KBPS == 800 || TX_KBPS == 615 || TX_KBPS == 500 || TX_KBPS == 400 || TX_KBPS == 250)
	ERROR value TX_KBPS not supported for "TX_KBPS"
 endif

#define PERIODE (8000/TX_KBPS)		;Tx periode (cycles)

	ORG	0
; R E S E T   E N T R Y 

	banksel	ANSELA
	clrf	ANSELA			;all pins = digital I/O	
	goto	mainreset

	ORG 4
;**********************************************************************************
;Interrupt service routine:
;int-source = PWM timer, 
;800 kbps: interrupt fires at rising edge of 7th clk pulse (T=0)
;(typ. execution time: 24 cycles)
interrupt
;T=3 (interrupt latency)
	banksel	0
 if (TX_KBPS == 800)			;ISR for 800 kbps data rate:
	bcf	PIR1,TMR2IF		
	btfsc	PR2,5			;just processed "reset" break? (long periode!)
	goto	autorestart		;y: maybe restart of transmission 
				 	
	movf	ptr_rd,W
	movwf	FSR0L
	subwf	ptr_wr,W		;no more data in buffer if ptr_rd == ptr_wr
	btfsc	STATUS,Z		;next byte available for transmission?
	goto	txstop			;no more data available: transmission finished
	incf	ptr_rd,F		;ptr_rd = next byte
;T=13! 
	clrf	TMR2			;adjust TMR2 to be in sync with SPI
	banksel SSP1STAT		
	movf	SSP1BUF,W		;T=16! dummy read to clear BF	
	movf	INDF0,W			;next data to transmit
	movwf	SSP1BUF			;restarts SPI transfer

	movf	bufsize,W
	subwf	ptr_rd,W
	btfsc	STATUS,Z		
	clrf	ptr_rd			;reached end of buffer, rollover to 0
	retfie

txstop				
	bcf	T2CON,TMR2ON		;stop TMR2 (PWM) after 8th pulse
	bcf	flags,TXRUN
	banksel MDCARL			;50s pause before next transmission:
	clrf	MDCARL			;carrier low = VSS: keep MD-output low
	movlw	53
	banksel	PR2			;set long PWM periode, so that 
	movwf	PR2			;next IRQ is after about 50 s
	bsf	T2CON,TMR2ON		;set TMR2 running
	retfie
 endif	;(TX_KBPS == 800)


 if (TX_KBPS == 615)			;ISR for 615 kbps data rate:
;T=4 
	bcf	PIR1,TMR2IF		
	btfsc	PR2,5			;just processed "reset" break? (long periode!)
	goto	autorestart		;y: maybe restart of transmission 
				 	
	movf	ptr_rd,W
	movwf	FSR0L
	subwf	ptr_wr,W		;no more data in buffer if ptr_rd == ptr_wr
	btfsc	STATUS,Z		;next byte available for transmission?
	goto	txstop			;no more data available: transmission finished
	incf	ptr_rd,F		;ptr_rd = next byte
;T=13! 
	movf	bufsize,W
	subwf	ptr_rd,W
	btfsc	STATUS,Z		
	clrf	ptr_rd			;reached end of buffer, rollover to 0
	clrf	TMR2			;adjust TMR2 to be in sync with SPI
	nop
	nop
	movf	INDF0,W			;next data to transmit
	banksel SSP1STAT		
	movwf	SSP1BUF			;restarts SPI transfer
	retfie
txstop	
	nop
	movlw	53
	bcf	flags,TXRUN
	bcf	T2CON,TMR2ON		;stop TMR2 (PWM) after 8th pulse
	banksel MDCARL			;50s pause before next transmission:
	clrf	MDCARL			;carrier low = VSS: keep MD-output low
	banksel	PR2			;set long PWM periode, so that 
	movwf	PR2			;next IRQ is after about 50 s
	bsf	T2CON,TMR2ON		;set TMR2 running
	retfie
 endif	;(TX_KBPS == 615)


 if (TX_KBPS == 500)
;500 kbps: interrupt fires at rising edge of 8th clk pulse
;T=4 
	btfsc	PR2,5			;just processed "reset" break? (long periode!)
	goto	autorestart		;y: maybe restart of transmission 
				 	
	movf	ptr_rd,W
	movwf	FSR0L
	subwf	ptr_wr,W		;no more data in buffer if ptr_rd == ptr_wr
	btfsc	STATUS,Z		;next byte available for transmission?
	goto	txstop			;no more data available: transmission finished
;T=11 
	banksel SSP1STAT		
	movf	INDF0,W			;next data to transmit
	movwf	SSP1BUF			;restarts SPI transfer

	banksel 0
	movlw	PERIODE/2+2
	movwf	TMR2			;adjust TMR2 to be in sync with SPI

	bcf	PIR1,TMR2IF		

	incf	ptr_rd,F		;ptr_rd = next byte
	movf	bufsize,W
	subwf	ptr_rd,W
	btfsc	STATUS,Z		
	clrf	ptr_rd			;reached end of buffer, rollover to 0
	retfie

txstop				
	movlw	48			;set long PWM periode, so that 
	movwf	PR2			;next IRQ is after about 50 s
	bcf	PIR1,TMR2IF		
	bcf	flags,TXRUN
	banksel MDCARL			;50s pause before next transmission:
	clrf	MDCARL			;carrier low = VSS: keep MD-output low
	retfie
 endif	;(TX_KBPS == 500)


 if (TX_KBPS == 400)
;400 kbps: interrupt fires at rising edge of 8th clk pulse
;T=4 
	btfsc	PR2,5			;just processed "reset" break? (long periode!)
	goto	autorestart		;y: maybe restart of transmission 
				 	
	movf	ptr_rd,W
	movwf	FSR0L
	subwf	ptr_wr,W		;no more data in buffer if ptr_rd == ptr_wr
	btfsc	STATUS,Z		;next byte available for transmission?
	goto	txstop			;no more data available: transmission finished
;T=11 
	banksel SSP1STAT		
	movf	INDF0,W			;next data to transmit
	movwf	SSP1BUF			;restarts SPI transfer

	banksel 0
	movlw	PERIODE/2+2
	movwf	TMR2			;adjust TMR2 to be in sync with SPI

	bcf	PIR1,TMR2IF		

	incf	ptr_rd,F		;ptr_rd = next byte
	movf	bufsize,W
	subwf	ptr_rd,W
	btfsc	STATUS,Z		
	clrf	ptr_rd			;reached end of buffer, rollover to 0
	retfie

txstop				
	movlw	48			;set long PWM periode, so that 
	movwf	PR2			;next IRQ is after about 50 s
	bcf	PIR1,TMR2IF		
	bcf	flags,TXRUN
	banksel MDCARL			;50s pause before next transmission:
	clrf	MDCARL			;carrier low = VSS: keep MD-output low
	retfie	
 endif	;(TX_KBPS == 400)


 if (TX_KBPS == 250)
;250 kbps: interrupt fires at rising edge of 8th clk pulse
;T=4 
	btfsc	PR2,5			;just processed "reset" break? (long periode!)
	goto	autorestart		;y: maybe restart of transmission 
				 	
	movf	ptr_rd,W
	movwf	FSR0L
	subwf	ptr_wr,W		;no more data in buffer if ptr_rd == ptr_wr
	btfsc	STATUS,Z		;next byte available for transmission?
	goto	txstop			;no more data available: transmission finished
;T=11 
	incf	ptr_rd,F		;ptr_rd = next byte
	movf	bufsize,W
	subwf	ptr_rd,W
	btfsc	STATUS,Z		
	clrf	ptr_rd			;reached end of buffer, rollover to 0

	banksel SSP1STAT		
	movf	INDF0,W			;next data to transmit
	movwf	SSP1BUF			;restarts SPI transfer

	banksel 0
	movlw	PERIODE/2+2
	movwf	TMR2			;adjust TMR2 to be in sync with SPI
	bcf	PIR1,TMR2IF		
	retfie

txstop				
	movlw	48			;set long PWM periode, so that 
	movwf	PR2			;next IRQ is after about 50 s
	bcf	PIR1,TMR2IF		
	bcf	flags,TXRUN
	banksel MDCARL			;50s pause before next transmission:
	clrf	MDCARL			;carrier low = VSS: keep MD-output low
	retfie
 endif	;(TX_KBPS == 250)


	
;50s pause is over, if auto-restart: initiate re-transmission of tx-buffer 
autorestart
	bcf	PIR1,TMR2IF		
	bcf	T2CON,TMR2ON		;stop TMR2 
	btfss	flags,ATXEN		;automatic restart enabled?
	retfie				;no, ISR complete

;automatic restart: begin transmission from start of buffer (addr. 0)					
	movlw	PERIODE-1		;set PWM periode for baud rate:
	movwf	PR2			;PWM periode = 4s, baudrate = 250kHz

	movlw	b'10000100'		;low carrier = CCP1 (PWM)
	banksel MDCARL
	movwf	MDCARL

	movlw	1
	movwf	ptr_rd			;send 2nd byte from buffer on next interrupt

	banksel	0 
	movlw	PERIODE/2+1
	movwf	TMR2
	movf	0x20,W			;tx first byte from buffer
	banksel SSP1CON1
	movwf	SSP1BUF  
	banksel	T2CON		
	bsf	T2CON,TMR2ON		;set TMR2 running
	bsf	flags,TXRUN
	retfie

;=============================================================== 
;Subroutine: starts transmission of data to WS2811
;(ptr_rd) = first data byte
start_tx
	movf	ptr_rd,W
	movwf	FSR0L

	incf	ptr_rd,F		;ptr_rd = next byte
	movf	bufsize,W	
	subwf	ptr_rd,W
	btfsc	STATUS,Z		
	clrf	ptr_rd			;reached end of buffer, rollover to 0

	banksel	T2CON
	btfsc	T2CON,TMR2ON		;wait if TMR2 is running
	goto	$-1			;(50 s break periode!)	
	bcf	PIR1,TMR2IF		

	movlw	PERIODE/2+1
	movwf	TMR2		
	movlw	PERIODE-1		
 	movwf	PR2

	movlw	b'10000100'		;low carrier = CCP1 (PWM)
	banksel MDCON
	movwf	MDCARL

	banksel SSP1CON1
	movf	SSP1BUF,W		;dummy read to clear BF	
	movf	INDF0,W
	movwf	SSP1BUF  
	banksel	T2CON			
	bsf	T2CON,TMR2ON		;set TMR2 running

	bsf	flags,TXRUN
	bsf	INTCON,PEIE
	bsf	INTCON,GIE
	banksel	gp_regs			;select standard bank
	return
	
;=============================================================== 
;Subroutine: send a single byte to tx-FIFO 
;does not check for FIFO overflow!
;temp contains data after return.
;total cycles (incl. call): 14
send2buf
	movwf	temp			;save data
	movf	ptr_wr,W
	movwf	FSR0L
	movf	temp,W
	movwi	FSR0++			;write data into buffer

	movf	bufsize,W
	subwf	FSR0L,W
	btfss	STATUS,Z		;case end of buffer: rollover to 0
	movf	FSR0L,W			;ptr++
	movwf	ptr_wr
	return

;=============================================================== 
;Subroutine: sends data for a single RGB-LED to tx-buffer
;input: FSR1 = 3 bytes of brightness data (WS2812: GRB)
;requirements: bufsize must be multiple of three, all prior write-ops
;to tx-buffer must write multiple of 3 bytes.
send_triple
	call	chkbuffer		;(T:10) wait if there is not enough space

	movf	ptr_wr,W		;FSR0 = indirect buffer access
	movwf	FSR0L			;(FSR0H = const = 0x20)

	moviw	0[FSR1]			;1st byte for current LED 
	movwi	FSR0++			;-> buffer

	moviw	1[FSR1]			;2nd byte for current LED 
	movwi	FSR0++			;-> buffer

	moviw	2[FSR1]			;3rd byte for current LED 
	movwi	FSR0++			;-> buffer

	movf	FSR0L,W
	movwf	ptr_wr			;adjust ptr_wr
	subwf	bufsize,W	
	btfsc	STATUS,Z		;ptr = end of buffer?
	call	wrap_ptr_wr		;y: wrap arround to 0, adjust system timer

	btfss	flags,TXRUN		;transmission active?
	call	start_tx		;no: initiate tx process

	banksel led_idx
	incfsz	led_idx,F
	goto	$+2
	incf	led_idx+1,F

	movlw	low CHAINLENGTH		;led_idx - CHAINLENGTH
	subwf	led_idx,W	
	movlw	high CHAINLENGTH
	subwfb	led_idx+1,W
	btfsc	STATUS,C
	bsf	flags,EOCHAIN		;>= 0: reached end of chain
	return


;=============================================================== 
;Subroutine: waits for free space (min. 3 bytes) in tx-buffer
;cycles (incl.call): 10 (if >= 3 bytes free) 
chkbuffer
	incf	ptr_wr,W		;check if there is enough free space in buffer (min 3 bytes):
	subwf	ptr_rd,W		;W = # of free bytes in buffer
	btfss	STATUS,C
	addwf	bufsize,W		;ptr_wr < ptr_rd: rollover!
	addlw	-3			;need space for 3 bytes  
	btfsc	STATUS,C
	return				;>= 3 bytes free!

chkbuf_wait
	call	sys_tm_adj		;buffer is full...
	btfss	flags,TXRUN		;just wait if transmission is running
	call	start_tx		;initiate tx
	goto	chkbuffer		

	
;***************************************************************************

wrap_ptr_wr
	clrf	ptr_wr

;---------------------------------------------------------------------------
;subroutine for updating sys_time, which is an infinite millisecond counter 
sys_tm_adj
	banksel	TMR0			;(TMR0 counts up every 32s)
	movf	sys_ts,W
	subwf	TMR0,W
	banksel	gp_regs
	btfss	WREG,7			;next ms timeout?
	return				;no, do nothing here!

	movlw	31			;reload ms-timer 
	addwf	sys_ts,F		;(31 * 32s = +992 s)

	movlw	3
	andwf	sys_time,W		;every 4th millisecond:
	btfsc	STATUS,Z		;add additional 32 s for correction
	incf	sys_ts,F		;(31+31+31+32 = 125, 125 * 32s = 4000s)

	incfsz	sys_time,F
	goto	sys_tm_adj		;case > 1 ticks passed since last call
	incf	sys_time+1,F
	goto	sys_tm_adj

;--------------------------------------------------------------------------

;**********************************************************************************
;       P R O G R A M   E N T R Y
;	Initialization of hardware:
mainreset
	banksel	OSCCON
	movlw	b'01110000'		;after reset, MPU is running on 500 kHz clock
	movwf	OSCCON			;set clock to 32MHz
 
	banksel	OSCSTAT
	btfss	OSCSTAT,PLLR
	goto	$-1			;wait if PLL is ready 
					;clock is now 32MHz
 
	clrf	FSR0H
	movlw	0x70
	movwf	FSR0L
initclr					;clr all vars (memory range 0x70..0x7F)
	clrf	INDF0
	incf	FSR0L,F
	btfss	FSR0L,7
	goto	initclr

 ifdef FIFOSIZE
	movlw	FIFOSIZE		;if FIFOSIZE is specified:
	movwf	bufsize			;set size!
 endif
	movlw	0x20			;from now on:
	movwf	FSR0H			;use FSR0 for linear buffer addressing only!
	clrf	FSR0L
clrbuf	
	clrf	INDF0			;clear led data buffer
	incf	FSR0L,F
 ifdef FIFOSIZE
	movf	bufsize,W
 else
	movlw	MAXBUFSIZE
	movwf	bufsize			;default: set bufsize to maximum
 endif
	subwf	FSR0L,W
	btfss	STATUS,Z
	goto	clrbuf

	banksel APFCON
	bsf	APFCON,7		;USART RX = RA5	
	bsf	APFCON,6		;SDO -> RA4 (comparator (-)input)

;--------------- Setup Timer 1
;	banksel	T1CON
;	movlw	b'00110001'		;T1 counting @1MHz
;	movwf	T1CON
;	clrf	TMR1L
;	clrf	TMR1H

;--------------- Setup Timer 0
	banksel	OPTION_REG
	movlw	b'11010111'		;TMR0 increments every 32s
	movwf	OPTION_REG			
 
;---------------- Setup PWM:
;PWM generates signal for 0-bit = 1/5 periode time high, 4/5 periode time low
	banksel PR2
	movlw	PERIODE-1		;PR2 value according to baudrate
	movwf	PR2
	banksel	SSP1ADD					
	movwf	SSP1ADD			;set SPI baudrate

	banksel CCP1CON
	movlw	0x0C
	movwf	CCP1CON			;PWM mode, P1A active high
				
	movlw	2			;pulse width for "0"-bit= 0.25s
	movwf	CCPR1L

	banksel	T2CON
	bcf	PIR1,TMR2IF
	clrf	TMR2
	bsf	T2CON,TMR2ON		;set TMR2 running

	btfss	PIR1,TMR2IF		;wait for one complete PWM cycle 
	goto	$-1
 if (TX_KBPS >= 615)
	movlw	b'00110000'		;stop timer, set postscaler 1:7
 else
	movlw	b'00111000'		;stop timer, set postscaler 1:8
 endif
	btfss	TMR2,2			;wait for TMR2 >4 (PWM out low)
	goto	$-1
	movwf	T2CON			
	bcf	PIR1,TMR2IF		;clr TMR2-interrupt flag
 	banksel	PIE1
	bsf	PIE1,TMR2IE		;enable TMR2 interrupt

;---------------- Setup SPI:
	banksel SSP1CON1		;(same bank for all SPI registers)
	movlw	b'00101010'		;SPI master, baudrate by SSP1ADD, idle CLK = low
	movwf	SSP1CON1
	bsf	SSP1STAT,CKE		;set clock edge: next data bit after CLK pulse

;---------------- Setup Comparator:
;Comparator delays SDO (= increase data hold time) in order to avoid glitches
	banksel FVRCON			 
	movlw	b'10001000'		;set voltage reference for comparator
	movwf	FVRCON			;(+)input to 2.048V

	banksel CM1CON0	
	movlw	b'10010100'		;enable, inv, high speed, no hysteresis, no sync
	movwf	CM1CON0			;no output to pin
	movlw	b'00100001'		;no interrupt, C1VP = FVR, C1VN = RA4
	movwf	CM1CON1 

;---------------- Setup Data Signal Modulator:
	banksel MDCON
	movlw	b'00000110'		;data source = comparator
	movwf	MDSRC
	movlw	b'10000001'		;high carrier = MDCIN1 (= SPICLK)
	movwf	MDCARH
;	movlw	b'10000100'		;low carrier = CCP1 (PWM)
	clrf	MDCARL			;low carrier = VSS until transmission starts
	movlw	b'11000000'		;enable modulator
	movwf	MDCON

;--------------- Setup Port A:
	banksel	PORTA
	clrf	PORTA
	banksel LATA
	clrf	LATA
	banksel TRISA
 ifdef SIMULATOR
	movlw	b'101010'
 else
	movlw	b'101100'		;RA0, RA1, RA4 = output, other ports = input/high Z
 endif
	movwf	TRISA	

	goto	user_main		;jump to user application

