;***************************************************************************
; Pattern Mover
;***************************************************************************

;--------------- initialize patternmover
patternmoverinit
	MOVLW_LIN pm_timer
	movwf	FSR0L
	movlw	10			;first action after 10 ms
	call	set_time_fsr0

 if N_LAYERS > 1
	movlw	N_LAYERS
	movwf	pp_layer
pm_inilp
	decf	pp_layer,W
	call	ctrldatfsr0
	movlw	0			;(does not affect Z-flag!)
	btfss	STATUS,Z		;skip if layer is not active
	movwi	pm_taccu[FSR0]		;clear "fraction of time" accumulator
	decfsz	pp_layer,F
	goto	pm_inilp
 else
	movlw	0
	call	ctrldatfsr0		;select control data 
	movlw	0
	btfss	STATUS,Z		;skip if layer is not active
	movwi	pm_taccu[FSR0]		;clear "fraction of time" accumulator
   ifdef USE_MICROSTEP
	btfss	STATUS,Z		;skip if layer is not active
	movwi	pdp_substep[FSR0]	;clear substep
   endif

 endif
	return


; calculates new starting locations for patterns on LED chain
; according to their speed data
; removes pattern from further processing (by 0 -> pp_data) if it has been moved
; out of chain 

patternmover
	MOVLW_LIN pm_timer
	call	check_time		;10 ms interval over?
	btfss	STATUS,C		
	return				;no

;------ 10 ms over: move the patterns...
	movlw	10			
	call	add_time_fsr0		;reload timer with another 10 ms

 if N_LAYERS > 1
	movlw	N_LAYERS
	movwf	pp_layer
pm_lloop
	decf	pp_layer,W	
 else
	movlw	0
 endif
	call	ctrldatfsr0
	btfsc	STATUS,Z
	goto	pm_nextlayer		;this layer is not in use

	call	startled2accu		;pp_accu = temporary 16 bit working register (startled)
	moviw	pm_speed[FSR0]
	movwf	temp
pm_fwdrpt
	moviw	pm_taccu[FSR0]
	btfsc	temp,7			;negative speed?
	goto	pm_movebak		;y move backwards

;speed >= 0:
	addwf	temp,F			;temp = pm_taccu + speed

	movlw	100/SPEEDFACTOR
	subwf	temp,W			;accumulated full step?
	btfss	STATUS,C		;(>= 100/SPEEDFACTOR?)
	movf	temp,W			;it's not a full step: pm_taccu += speed
	movwi	pm_taccu[FSR0]	

	btfss	STATUS,C
	goto	pm_finished		;no, don't move pattern

	call	pm_forwardstep
	clrf	temp
	goto	pm_fwdrpt		;in case of multiple steps are needed!


;---------- speed <0: move pattern backwards
pm_movebak
	addwf	temp,W			;pm_accu + (-)speed -> W	
	movwi	pm_taccu[FSR0]		;case it's not a full step: pm_taccu += speed
	btfsc	STATUS,C
	goto	pm_finished		;not yet a full step, don't move pattern

pm_bakrpt
	addlw	100/SPEEDFACTOR
	movwi	pm_taccu[FSR0]		;full step: pm_accu += 100/SPEEDFACTOR + speed

	call	pm_backstep		;move pattern in reverse direction by one step

	moviw	pm_taccu[FSR0]
	btfsc	WREG,7			;taccu still <0?
	goto	pm_bakrpt		;y: perform another step back

;---------- 	
pm_finished
	call	accu2startled		;store new startled in layer's data set

pm_nextlayer
 if N_LAYERS > 1
	decfsz	pp_layer,F		;next layer...
	goto	pm_lloop
 endif
	goto	patternmover		;in case >1 time steps to be executed

;---------- move pp_accu -> pdp_startled[FSR0]
accu2startled
	movf	pp_accu,W
	movwi	pdp_startled[FSR0]
	movf	pp_accu+1,W
	movwi	pdp_startled+1[FSR0]
	return

;---------- move pdp_startled[FSR0] -> pp_accu
startled2accu
	moviw	pdp_startled[FSR0]
	movwf	pp_accu			;temporary 16 bit working register = startled
	moviw	pdp_startled+1[FSR0]
	movwf	pp_accu+1
	return

;----------- Subroutine: move pattern forward by one step 
pm_forwardstep
 ifdef USE_MICROSTEP
	call	checkustepmode		;microstepping?
	btfss	STATUS,Z
	goto	pm_fwd_skipustep	;no, skip

	moviw	pdp_substep[FSR0]
	addlw	1
	andlw	7			;rollover 7 -> 0
	movwi	pdp_substep[FSR0]
	btfss	STATUS,Z
	return				;just moving by a substep
pm_fwd_skipustep
 endif

	incfsz	pp_accu,F		;startled++
	goto	$+2
	incf	pp_accu+1,F	

;check for >= CHAINLENGTH
	btfsc	pp_accu+1,7
	return				;startled <0 

	movlw	low CHAINLENGTH		;startled - CHAINLENGTH
	subwf	pp_accu,W
	movlw	high CHAINLENGTH
	subwfb	pp_accu+1,W
	btfss	STATUS,C
	return				;startled < CHAINLENGTH

;pattern moved beyond end of chain:
	btfss	INDF0,WRAPEN		;wrapping enabled?
	goto	pm_disable		;no wrapping -> pattern is out of display-range

	btfss	INDF0,TILE		
	goto	pm_frestart		;no tile mode: rollover to start of chain

;tile & wrapping: start of pattern might not fit with start of chain,
;restore startled by length of pattern (= *3 LEDs)
	call	pp_getpatlen		;length of pattern -> w16
	movf	w16,W
	subwf	pp_accu,F		;startled - w16 -> startled
	movf	w16+1,W
	subwfb	pp_accu+1,F
	return
 
pm_disable
	movlw	0x20
	movwf	FSR1H
	MOVLW_LIN pp_data		;disable further processing...
	addwf	pp_layer,W
	movwf	FSR1L
	clrf	INDF1			;pp_data[layer] = 0
	return

pm_frestart
	clrf	pp_accu
	clrf	pp_accu+1
	return

 ifdef USE_MICROSTEP
;----------- Subroutine: check for microstepping blend mode  
checkustepmode
	moviw	pdp_blendmode[FSR0]
	xorlw	BM_UST_OPAQUE		
	btfsc	STATUS,Z
	return
	moviw	pdp_blendmode[FSR0]
	xorlw	BM_UST_ADD		
	return
 endif


;----------- Subroutine: move pattern backwards by one step 
pm_backstep
 ifdef USE_MICROSTEP
	call	checkustepmode		;microstepping?
	btfss	STATUS,Z
	goto	pm_back_skipustep	;no, skip

	moviw	pdp_substep[FSR0]
	addlw	-1
	btfss	STATUS,C
	movlw	7			;rollover 0 -> 7
	movwi	pdp_substep[FSR0]
	btfsc	STATUS,C
	return				;just moving by a substep
pm_back_skipustep
 endif
	movlw	-1
	addwf	pp_accu,F		;startled--
	addwfc	pp_accu+1,F

	btfss	pp_accu+1,7			
	return				;startled >= 0

	call	pp_getpatlen		;length of pattern -> w16

	btfss	INDF0,WRAPEN
	goto	pm_bsnowrap		;no wrapping
	btfsc	INDF0,TILE		
	goto	pm_bswraptile		;wrap + tile

;wrapping, but no tile: wrap around 0 -> end of chain
	movlw	low (CHAINLENGTH-1)
	movwf	pp_accu
	movlw	high (CHAINLENGTH-1)
	movwf	pp_accu+1
	return

pm_bswraptile
	movf	w16,W
	addwf	pp_accu,F		;startled += patternlength
	movf	w16+1,W
	addwfc	pp_accu+1,F
	return

pm_bsnowrap
	call	accu2startled
	btfsc	INDF0,TILE
	goto	pm_bstile		;tile mode

;standard mode: disable processing if pattern has completely moved out of chain
	call	start_led_x3		;start_led * 3 -> led_idx (negative!)
	movf	led_idx,W
	addwf	w16,W
	movf	led_idx+1,W
	addwfc	w16+1,W
	btfss	STATUS,C
	goto	pm_disable
	return

pm_bstile
	call	pp_prepare		;let this subroutine check for ready-condition
	call	startled2accu
	btfsc	INDF0,DONE
	goto	pm_disable
	return

	


	



