;***************************************************************************
; Pattern Processor
;***************************************************************************

;Generate labels for pp_ctrl[layer] data structures:
 variable i=0
 while i<N_LAYERS
 LIN2TDM PP_CTRL0_LIN + i*PATTBLKLEN
pp_ctrl#v(i) equ BNK_ADDR	
i+=1
 endw

;--------------------------------------------------------------------------
; FSR0 = pattern control data structure (layer: W) 
; returns STATUS.Z = 1 if there are no pattern data for this layer
ctrldatfsr0
	TDM2LIN	pp_data
	addlw	LIN_ADDR		;address + layer
	movwf	FSR0L			
	movf	INDF0,W			;get pointer from table (pp_data[layer])
	movwf	FSR0L			;FSR = points to data 
	return

;--------------------------------------------------------------------------
; assembles LED data from source patterns and transmits data for all active
; layers to entire LED chain. 

patternprocessor
 ifdef USE_PATMOVER
	call	patternmover		;adjust patterns start positions
 endif
	banksel	gp_regs
	bcf	flags,ATXEN		;no automatic restart (for ISR)!
;automatically adjusts buffer size in order to keep FIFO data below variables:
	movf	bufsize,W		;bufsize >= pp_ctrlblk0_lin? 
	sublw	PP_CTRL0_LIN		;first register RAM location - bufsize
	btfsc	STATUS,C		;(FIFO would overwrite registers!)
	goto	pp_bufsizeok		;no, bufsize seems to be valid

	movlw 	3*(PP_CTRL0_LIN/3)	;must be multiple of 3
	movwf	bufsize			;set bufsize to keep FIFO below variable storage 
	clrf	ptr_rd			;for sure...
	clrf	ptr_wr			

pp_bufsizeok	
	movlw	N_LAYERS
	movwf	pp_layer			;loop counter through all layers

;-------- prepare pattern data according to mode (tile, wrap) and startled:
pp_preplp
	decf	pp_layer,W		;layer-# (0..n) -> W
	call	ctrldatfsr0		;FSR0 = pointer to ctrl data
	btfsc	STATUS,Z
	goto	pp_prepskip		;ptr= 0: layer not active!

	call 	pp_prepare		;prepare pattern data


pp_prepskip
	decfsz	pp_layer,F			;next layer...
	goto	pp_preplp
;---------

	clrf	led_idx			;++ in each loop
	clrf	led_idx+1

	movlw	low (-CHAINLENGTH)	
	movwf	pp_accu			;(loop counter for fast check of loop condition)
	
	btfss	flags,TXRUN		;previous output complete?
	goto	patproclp		;y: can write to FIFO regularly

;--------- Transmission of previous chain data is still going on 
;Uses alternate write pointer in order to avoid system blocking.
;Meanwhile, ISR can transmit data from FIFO until it detects 
;regular "buffer empty", and CPU time is not wasted by
;stupid "waiting for transmission complete".
 
pp_altwrptr
	movf	ptr_wr,W		;switch to alternate wr-pointer
	movwf	ptr_wr_alt
;==================================================================
;loop for processing all LEDs in chain 
patproclp_alt
	call	pixel_all_layers	;data -> pp_red, pp_grn, pp_blue

pp_altlp1	
	btfsc	flags,TXRUN		;transmission still active?
	goto	pp_altlp2		;y: continue using alternate wr-pointer

	movf	ptr_wr_alt,W		;(tx break) complete buffer was sent,
	movwf	ptr_wr			;switch to regular write pointer
	goto	patproclp2
		
pp_altlp2
	incf	ptr_wr_alt,W		;check if there is enough free space in buffer (min 3 bytes):
	subwf	ptr_rd,W		;W = # of free bytes in buffer
	btfss	STATUS,C
	addwf	bufsize,W		;ptr_wr < ptr_rd: rollover!
	addlw	-3			;need space for 3 bytes  
	btfss	STATUS,C
	goto	pp_altlp1		;< 3 bytes free, wait!

	movf	ptr_wr_alt,W		;FSR0 = indirect buffer access
	movwf	FSR0L			;(FSR0H = const = 0x20)

	movf	pp_grn,W		;1st byte for current LED 
	movwi	FSR0++			;-> buffer

	movf	pp_red,W		;2nd byte for current LED 
	movwi	FSR0++			;-> buffer

	movf	pp_blue,W		;3rd byte for current LED 
	movwi	FSR0++			;-> buffer

	movf	FSR0L,W
	movwf	ptr_wr_alt		;adjust alternate wr-pointer
	subwf	bufsize,W	
	btfss	STATUS,Z		;ptr = end of buffer?
	goto	pp_altlp3		;no

	clrf	ptr_wr_alt		;y: wrap arround to 0
	call	sys_tm_adj

pp_altlp3
	incfsz	led_idx,F		;++ 
	goto	$+2
	incf	led_idx+1,F	

	incfsz	pp_accu,F		;(fast loop counter, in sync with led_idx)
	goto	patproclp_alt		;not yet end of chain: process next LED

	movlw	high CHAINLENGTH
	xorwf	led_idx+1,W
	btfss	STATUS,Z
	goto	patproclp_alt		;not yet end of chain: next LED	

;execution of this code can only happen with short LED chains, whose complete
;data fit into FIFO: 
	btfsc	flags,TXRUN
	goto	$-1			;wait if still transmitting previous data 
	movf	ptr_wr_alt,W		
	movwf	ptr_wr			;switch to regular write pointer
	call	start_tx		;start tx of current chain data
	goto	sys_tm_adj

;==================================================================
;loop for processing all LEDs from chain 
patproclp
	call	pixel_all_layers	;data -> pp_grn, pp_red, pp_blue

patproclp2
;sends LED colour data to tx-buffer
;	call	chkbuffer		;(T:10) wait if there is not enough space
	CHKBUFFER			;(same function as call, T:7)
					;(initiates transmission if buffer is full)
	movf	ptr_wr,W		;FSR0 = indirect buffer access
	movwf	FSR0L			;(FSR0H = const = 0x20)

	movf	pp_grn,W		;1st byte for current LED 
	movwi	FSR0++			;-> buffer

	movf	pp_red,W		;2nd byte for current LED 
	movwi	FSR0++			;-> buffer

	movf	pp_blue,W		;3rd byte for current LED 
	movwi	FSR0++			;-> buffer

	movf	FSR0L,W
	movwf	ptr_wr			;adjust ptr_wr
	subwf	bufsize,W	
	btfsc	STATUS,Z		;ptr = end of buffer?
	call	wrap_ptr_wr		;y: wrap arround to 0, adjust system timer

	incfsz	led_idx,F
	goto	$+2
	incf	led_idx+1,F

	incfsz	pp_accu,F		;(loop counter, fast detection of loop condition)
	goto	patproclp		;not yet end of chain: process next LED
	movlw	high CHAINLENGTH
	xorwf	led_idx+1,W
	btfss	STATUS,Z
	goto	patproclp		;not yet end of chain: process next LED

;if data of whole chain fit into FIFO, transmission might still be not initiated:
	btfss	flags,TXRUN		;transmission active?
	goto	start_tx		;no: initiate tx process
	return



;=================================================================================
;Subroutine: length of pattern (incl. gap) -> w16, srcptr = start of pattern
pp_getpatlen
	clrf	w16
	clrf	w16+1
	btfss	INDF0,TILE
	goto	pp_gpl_nogap		;no tile: no gap!

	moviw	pdp_gap+1[FSR0]		;gap (in pixels!)
	movwf	w16+1
	moviw	pdp_gap[FSR0]
	movwf	w16
	lslf	w16,F			; * 3 = gap in # of data bytes
	rlf	w16+1,F
	addwf	w16,F
	moviw	pdp_gap+1[FSR0]		
	addwfc	w16+1,F			;w16 = length of gap (in data bytes)

pp_gpl_nogap
	moviw	pdp_srcend[FSR0]
	addwf	w16,F	
	moviw	pdp_srcend+1[FSR0]
	addwfc	w16+1,F	
	moviw	pdp_srcstart[FSR0]
	movwi	pdp_srcptr[FSR0]	;(case start of pattern == start of chain)
	subwf	w16,F			;W16 = length of data (pattern + gap)
	moviw	pdp_srcstart+1[FSR0]	
	movwi	pdp_srcptr+1[FSR0]
	subwfb	w16+1,F	
	return	

;=================================================================================
;Subroutine: start_led * 3 -> led_idx
start_led_x3
	moviw	pdp_startled+1[FSR0]
	movwf	led_idx+1
	moviw	pdp_startled[FSR0]
	movwf	led_idx	
	aslf	led_idx,F
	rlf	led_idx+1,F
	addwf	led_idx,F	
	moviw	pdp_startled+1[FSR0]
	addwfc	led_idx+1,F
	return

;=================================================================================
;Subroutine: led_idx += srcstart, then copy led_idx -> srcptr 
;GAPFLG=1 if srcptr >= end of source pattern 
pp_addsrcstart
	moviw	pdp_srcstart[FSR0]
	addwf	led_idx,F
	movf	led_idx,W
	movwi	pdp_srcptr[FSR0]
	moviw	pdp_srcstart+1[FSR0]
	addwfc	led_idx+1,F	
	movf	led_idx+1,W	
	movwi	pdp_srcptr+1[FSR0]

	moviw	pdp_srcend[FSR0]
	subwf	led_idx,W		;ptr - srcend: ptr > end of pattern?
	moviw	pdp_srcend+1[FSR0]
	subwfb	led_idx+1,W
	bcf	flags,GAPFLG
	btfsc	STATUS,C		;C=1 if	ptr > active pattern area
	bsf	flags,GAPFLG		; C -> GAPFLG
	return

;=================================================================================
;Subroutine: negate led_idx
neg_idx
	comf	led_idx,F		
	comf	led_idx+1,F
	incfsz	led_idx,F
	return
	incf	led_idx+1,F
	return	

;==========================================
;led_idx/3 -> led_idx
;(used: w16, pp_accu)
idx_div3
	clrf	w16+1
	movlw	3
	movwf	w16
	call	idx_mod_w16
	movf	pp_accu,W
	movwf	led_idx
	movf	pp_accu+1,W
	movwf	led_idx+1
	return

;==========================================
;16 bit division & modulo subroutine:
;led_idx mod w16 -> led_idx
;led_idx/w16 -> pp_accu (remainder in led_idx)
;flags: Z=1 if result is zero
;w16: no change

idx_mod_w16
	movf	led_idx,W
	movwf	pp_accu
	movf	led_idx+1,W
	movwf	pp_accu+1
	clrf	led_idx
	clrf	led_idx+1

	movlw	16
	movwf	temp		;loop: 16 shifts
mod_lp
	rlf	pp_accu,F
	rlf	pp_accu+1,F
	rlf	led_idx,F
	rlf	led_idx+1,F

	movf	w16,W
	subwf	led_idx,W
	movf	w16+1,W
	subwfb	led_idx+1,W
	btfss	STATUS,C
	goto	mod_lpnext	;led_idx < w16: next loop

	movwf	led_idx+1	;led_idx >= w16: led_idx -= w16
	movf	w16,W
	subwf	led_idx,F
	bsf	STATUS,C

mod_lpnext
	decfsz	temp,F
	goto	mod_lp

	rlf	pp_accu,F	;shift last bit into result
	rlf	pp_accu+1,F

	movf	led_idx,W	;return Z=1 if remainder == 0
	iorwf	led_idx+1,W
	return

;=================================================================================
;Subroutine for preparing pattern data
pp_prepare
	bcf	INDF0,ACTIVE		;default flag setting: pattern not active
	bcf	INDF0,WRAPFLG	

	moviw	pdp_startled+1[FSR0]
	btfss	WREG,7			;startposition of pattern negative?
	goto	pp_prep1		;startpos. >= 0

;startpos. of pattern < 0  -> check if end of pattern extends into chain 
	bsf	INDF0,DONE		;default state: done
	call	start_led_x3		;start_led * 3 -> led_idx (negative!)
	call	neg_idx			;change to positive val.

	call	pp_addsrcstart		;srcptr = start of pattern + led_idx
	btfss	flags,GAPFLG
	goto	pp_intochain_actv	;< end of pattern: pt.extends into chain

	btfss	INDF0,TILE		;tile mode?
	return				;no, single pattern is out of range -> invisible!

	call	pp_getpatlen		;length of pattern -> w16, srcptr = start of pattern
	call	start_led_x3		;start_led * 3 -> led_idx (negative!)
	call	neg_idx			;change to positive
	call	idx_mod_w16		;led_idx (= # of bytes till LED0) modulo w16
	call	pp_addsrcstart		;srcptr = start of pattern + led_idx

	btfsc	INDF0,WRAPEN		;wrapping enabled?
	goto	pp_intochain		;y: distribute pattern over full chain

;no wrapping, tiled patterns should end at the end of chain + startled (startled <0)
;(w16 still contains length of pattern+gap in bytes) 
	moviw	pdp_srcstart[FSR0]	;led_idx = # of bytes into pattern
	subwf	led_idx,F
	moviw	pdp_srcstart+1[FSR0]
	subwfb	led_idx+1,F

	moviw	pdp_startled+1[FSR0]	;pp_accu = startled*3
	movwf	pp_accu+1		;(=byte distance LED0->startled)
	moviw	pdp_startled[FSR0]
	movwf	pp_accu
	lslf	pp_accu,F
	rlf	pp_accu+1,F
	addwf	pp_accu,F
	moviw	pdp_startled+1[FSR0]
	addwfc	pp_accu+1,F

	movf	led_idx,W		;pp_accu = byte distance: beginning of pattern->startled
	addwf	pp_accu,F
	movf	led_idx+1,W
	addwfc	pp_accu+1,F

	btfss	flags,GAPFLG
	goto	pp_cmplength		;LED0 not within gap

	movf	w16,W
	subwf	pp_accu,F		;LED0 within gap: compare start of next tile
	movf	w16+1,W
	subwfb	pp_accu+1,F

pp_cmplength
	movlw	low (3*(CHAINLENGTH-1))
	addwf	pp_accu,W
	movlw	high (3*(CHAINLENGTH-1))
	addwfc	pp_accu+1,W
	btfsc	WREG,7
	return				;(chainlengh + begin of patt.) < 0, no display at all	

pp_intochain
	btfss	flags,GAPFLG		;skip if ptr >= end of active area
	goto	pp_intochain_actv	;ptr < end of active pattern area: 
					;visible part of pattern extends into chain

;compute LED number, where next visible part of pattern begins:
pp_procgap
	moviw	pdp_srcptr[FSR0]
	movwf	led_idx
	moviw	pdp_srcptr+1[FSR0]
	movwf	led_idx+1

	moviw	pdp_srcend[FSR0]
	subwf	led_idx,F		;srcptr - srcend -> led_idx
	moviw	pdp_srcend+1[FSR0]	
	subwfb	led_idx+1,F		; = length into gap (bytes)	
	call	idx_div3		; /3 = number of pixels into gap

	moviw	pdp_gap[FSR0]
	subwf	led_idx,F		;(pixels into gap) - (length of gap) 
	moviw	pdp_gap+1[FSR0]
	subwfb	led_idx+1,F		; = -(remaining length of gap)
	call	neg_idx			; = (remaining length of gap)
	movf	led_idx,W
	movwi	pdp_srcptr[FSR0]	;triggers ACTIVE if led_idx == pdp_srcptr
	movf	led_idx+1,W
	movwi	pdp_srcptr+1[FSR0]
	goto	$+2			;skip setting ACTIVE-flag

pp_intochain_actv
	bsf	INDF0,ACTIVE
	bcf	INDF0,DONE
	return
	

;startled >= 0:
pp_prep1
	bcf	INDF0,DONE
	btfsc	INDF0,WRAPEN		;wrapping enabled?
	goto	pp_prep_wrapping	;y

;no wrapping:
pp_set_trigstart			
	moviw	pdp_startled[FSR0]	;set trigger (ACTIVE) = startled
	movwi	pdp_srcptr[FSR0]	
	moviw	pdp_startled+1[FSR0]
	movwi	pdp_srcptr+1[FSR0]
	return
	
;wrapping enabled:
pp_prep_wrapping
	btfss	INDF0,TILE		;both, WRAPEN + TILE?
	goto	pp_checkwrap		;just wrapping: check if pattern exceeds LED-chain

;------ both, wrap and tile mode are active: get source address for first LED's data
	call	pp_getpatlen		;length of pattern (incl.gap) -> w16, srcptr = start of pattern
	call	start_led_x3		;start_led * 3 -> led_idx
	call	idx_mod_w16		;led_idx modulo w16
	btfsc	STATUS,Z
	goto	pp_prepsetactv		;pattern start matches with start of LED chain
	
	call 	neg_idx	
	movf	w16,W			;length (pattern + gap)
	addwf	led_idx,F		;(length - led_idx -> led_idx)
	movf	w16+1,W
	addwfc	led_idx+1,F
	call	pp_addsrcstart		;srcptr = start of pattern + led_idx
	btfss	flags,GAPFLG
	goto	pp_prepsetactv		;srcptr is within active pattern area
	goto	pp_procgap		;srcptr is in range of gap		
;------

;------ processes wrap-arround if part of pattern exceeds the end of LED-chain 
;(wrap=on, tile=off)
pp_checkwrap
	moviw	pdp_startled[FSR0]
	sublw	low CHAINLENGTH
	movwf	w16			; diff = startled till end of chain 
	movwf	led_idx

	moviw	pdp_startled+1[FSR0]
	btfss	STATUS,C
	addlw	1			;borrow: sub one more...
	sublw	high CHAINLENGTH
	movwf	w16+1
	movwf	led_idx+1

	lslf	w16,W			;= diff. *2
	rlf	w16+1,F		
	addwf	led_idx,F
	movf	w16+1,W
	addwfc	led_idx+1,F		;diff. * 3 = # of bytes (pt_start..end_of_chain) 

	call	pp_addsrcstart		;srcptr = led_idx + srcstart
	btfsc	flags,GAPFLG
	goto	pp_set_trigstart	;ptr >= srcend: end of pattern within LED-chain, no wrapping
		
	bsf	INDF0,WRAPFLG		;avoids setting "DONE" @end of pattern
pp_prepsetactv
	bsf	INDF0,ACTIVE		;active pattern area
	return


;--------------------------------------------------------------------------
;process data for a single RGB-LED in all layers
;output => pp_red, pp_grn, pp_blue

pixel_all_layers
 if TX_KBPS <615
	movlw	b'00011111'		;every 32th LED:
	andwf	led_idx,W
	btfsc	STATUS,Z	
	call	sys_tm_adj		;maintain system timer
 endif
;---- set background:
 ifdef bckgr_grn
	movf	bckgr_grn,W
	movwf	pp_grn			;initialize pixel to background colour
 else	
	clrf	pp_grn			;initialize pixel to "off"
 endif
 ifdef bckgr_red
	movf	bckgr_red,W
	movwf	pp_red			;initialize pixel to background colour
 else	
	clrf	pp_red			;initialize pixel to "off"
 endif
 ifdef bckgr_blue
	movf	bckgr_blue,W
	movwf	pp_blue			;initialize pixel to background colour
 else	
	clrf	pp_blue			;initialize pixel to "off"
 endif
;---- process layers:
;==== while loop generates code for all layers:
 variable i=0
 while i < N_LAYERS

	movf	pp_data+i,W		;layer 0..n
	btfss	STATUS,Z		;skip if layer is off
i += 1
 if i<N_LAYERS
	call	pattlayer		;(>=34)
 else
	goto	pattlayer
 endif

 endw
	return





;pattern is not active, check if it should be activated:
pp_chk2
	moviw	pdp_srcptr[FSR0]	;check current position: start of pattern here?
	xorwf	led_idx,W		;led_idx == startindex?
	btfss	STATUS,Z
	return				;no, keep LED colour from underlaying layers
	moviw	pdp_srcptr+1[FSR0]
	xorwf	led_idx+1,W
	btfss	STATUS,Z
	return		

;match of start-index: triggers sending of pattern data! 	
	bsf	INDF0,ACTIVE		;set activity flag
	moviw	pdp_srcstart[FSR0]	
	movwf	FSR1L			;FSR1 = source pattern startaddr.
	moviw	pdp_srcstart+1[FSR0]
	movwf	FSR1H
	goto	pp_rdpixel		;process first pixel...


;--------------------------------------------------------------------------
;process single pixel from selected pattern (W = address of pattern data structure)
pattlayer
	movwf	FSR0L			;FSR0 = ptr. to pattern structure
	btfsc	INDF0,DONE		
	return				;pattern has already been processed

	btfss	INDF0,ACTIVE		
	goto	pp_chk2			;pattern is currently not active

;pattern is active: set FSR1 = source data
	moviw 	pdp_srcptr[FSR0]
	movwf	FSR1L	
	moviw 	pdp_srcptr+1[FSR0]
	movwf	FSR1H	

pp_rdpixel
	moviw	pdp_blendmode[FSR0]	;get blend mode (= branch distance!)
	brw
pp_bmref				;(label for reference)				
;--------------- pixel processing(BM_OPAQUE) entry point: 
;underlaying LED-colour is replaced by pattern data
pp_opaque				
	moviw	FSR1++
	movwf	pp_grn
	moviw	FSR1++
	movwf	pp_red
	moviw	FSR1++
	movwf	pp_blue
	goto	pp_ptradj

;--------------- pixel processing(BM_SUBTRACT) entry point: 
;underlaying LED-brightness is reduced by pattern data
pp_subtract			
	moviw	FSR1++
	subwf	pp_grn,F
	btfss	STATUS,C
	clrf	pp_grn

	moviw	FSR1++
	subwf	pp_red,F
	btfss	STATUS,C
	clrf	pp_red

	moviw	FSR1++
	subwf	pp_blue,F
	btfss	STATUS,C
	clrf	pp_blue
	goto	pp_ptradj

;--------------- pixel processing(BM_MIX) entry point: 
;underlaying LED-colour is mixed with pattern data (50%)
pp_mix				
	moviw	FSR1++
	addwf	pp_grn,F
	rrf	pp_grn,F

	moviw	FSR1++
	addwf	pp_red,F
	rrf	pp_red,F

	moviw	FSR1++
	addwf	pp_blue,F
	rrf	pp_blue,F
	goto	pp_ptradj

 ifdef USE_MICROSTEP
;--------------- microstep processing, opaque 
pp_ustep_opaque
	call	ustep_sub		;(T_avrg. of subroutine: 37)
	movf	temp_grn,W
	movwf	pp_grn	
	movf	temp_red,W
	movwf	pp_red	
	movf	temp_blue,W
	movwf	pp_blue	
	goto	pp_ptradj

;--------------- microstep processing, add 
pp_ustep_add
	call	ustep_sub		;(T_avrg. of subroutine: 37)
	movf	temp_grn,W
	addwf	pp_grn,W
	btfsc	STATUS,C
	movlw	0xFF
	movwf	pp_grn

	movf	temp_red,W
	addwf	pp_red,W
	btfsc	STATUS,C
	movlw	0xFF
	movwf	pp_red

	movf	temp_blue,W
	addwf	pp_blue,W
	btfsc	STATUS,C
	movlw	0xFF
	movwf	pp_blue
	goto	pp_ptradj

 endif

;--------------- pixel processing(BM_MULTIPLY) entry point: 
;underlaying LED-colour is multiplied with current pattern data
pp_multiply
	moviw	FSR1++
	movwf	temp_grn
	moviw	FSR1++
	movwf	temp_red
	moviw	FSR1++
	movwf	temp_blue
	call	pp_ptradj	;adjust srcptr
	goto	pp_mulsub	;pp_<col> = pp_<col> * temp_<col> 


;--------------- pixel processing(BM_ADD) entry point: 
;pattern data are added to underlaying LED-colour
pp_add								
	moviw	FSR1++
	addwf	pp_grn,W
	btfsc	STATUS,C
	movlw	0xFF
	movwf	pp_grn

	moviw	FSR1++
	addwf	pp_red,W
	btfsc	STATUS,C
	movlw	0xFF
	movwf	pp_red

	moviw	FSR1++
	addwf	pp_blue,W
	btfsc	STATUS,C
	movlw	0xFF
	movwf	pp_blue

;---------------- all blend modes come here again:
pp_ptradj
	movf	FSR1L,W	
	movwi 	pdp_srcptr[FSR0]	;adjust pointer for reading next pattern data
	movf	FSR1H,W	
	movwi 	pdp_srcptr+1[FSR0]

	moviw	pdp_srcend[FSR0]	
	subwf	FSR1L,W			;FSR1 == src_end: reached end of pattern?
	btfss	STATUS,Z
	return				;not yet end of pattern
	
	moviw	pdp_srcend+1[FSR0]
	subwfb	FSR1H,W			
	btfss	STATUS,Z
	return				;not yet end of pattern

;reached end of pattern:
pp_endpat
	btfsc	INDF0,TILE		;tile mode?
	goto	pp_tile			;y: rewind ptr to start of pattern data

	moviw 	pdp_startled[FSR0]
	movwi 	pdp_srcptr[FSR0]	;reload trigger: 
	moviw 	pdp_startled+1[FSR0]	;display restarts if led_idx == pdp_srcptr
	movwi 	pdp_srcptr+1[FSR0]

	btfss	INDF0,WRAPFLG		;was wrapped part of pattern at start of chain?
pp_endpat1
	bsf	INDF0,DONE		;no: all done for now...
	bcf	INDF0,WRAPFLG
	bcf	INDF0,ACTIVE
	return 

	
pp_tile	
	moviw	pdp_srcstart[FSR0]	;case tile mode without gap: 	
	movwi	pdp_srcptr[FSR0]	;repeat pattern data from start immediately...
	moviw	pdp_srcstart+1[FSR0]		
	movwi	pdp_srcptr+1[FSR0]

	moviw	pdp_gap[FSR0]
	btfss	STATUS,Z	
	goto	pp_tilegap		;gap != 0: process gap
	moviw	pdp_gap+1[FSR0]
	btfsc	STATUS,Z	
	goto	pp_tile_nogap		;no gap
	
pp_tilegap
	moviw	pdp_gap[FSR0]
	bsf	STATUS,C		;led_idx not yet incr.: +1
	addwfc	led_idx,W
	movwi	pdp_srcptr[FSR0]	;reload trigger: 
	moviw	pdp_gap+1[FSR0]
	addwfc	led_idx+1,W
	movwi	pdp_srcptr+1[FSR0]	;led_idx + gap
	bcf	INDF0,ACTIVE

pp_tile_nogap
	btfsc	INDF0,WRAPEN
	return				;wrapping enabled: distribute pattern over whole chain

	moviw	pdp_startled+1[FSR0]
	btfss	WREG,7
	return				;startled is >= 0

;no wrapping, start position negative: keep end of chain clear
	movwf	w16+1
	moviw	pdp_startled[FSR0]
	addlw	low (CHAINLENGTH-1)
	movwf	w16
	movlw	high (CHAINLENGTH-1)
	addwfc	w16+1,F			;w16 = chainlength-1 + startled (startled is negative!)

	moviw	pdp_gap[FSR0]
	subwf	w16,F			
	moviw	pdp_gap+1[FSR0]
	subwfb	w16+1,F			;w16 -= gap
	
	btfsc	w16+1,7
	goto	pp_endpat1		;<0: no display

	movf	w16,W
	subwf	led_idx,W		;led_idx >= w16?
	movf	w16+1,W
	subwfb	led_idx+1,W
	btfsc	STATUS,C
	goto	pp_endpat1		;y: rest of chain = off
	return				;n: continue displaying tiled pattern
