;***************************************************************************
; Microstep Interpolation
;***************************************************************************


;--------------- microstep processing subroutine
;average running time (incl. call+return): 37 cycles 
ustep_sub
	moviw	-3[FSR1]
	movwf	temp_grn	;grn[n-1]
	moviw	-2[FSR1]
	movwf	temp_red	;red[n-1]
	moviw	-1[FSR1]
	movwf	temp_blue	;blue[n-1]

	moviw	pdp_substep[FSR0]
	brw
	goto	ustep0		;substep = 0: nothing to interpolate (weight: 100%[n])
	goto	ustep1		;weight: 1/8[n-1] + 7/8[n]
	goto	ustep2
	goto	ustep3
	goto	ustep4
	goto	ustep5
	goto	ustep6

;pattern pos. +7/8: weight is 1/8 current + 7/8 previous
ustep7
	moviw	FSR1++
	addwf	temp_grn,W
	rrf	WREG,W		;W = 1/2 grn[n-1] + 1/2 grn[n] 
	addwf	temp_grn,W
	rrf	WREG,W		;W = 3/4 grn[n-1] + 1/4 grn[n]
	addwf	temp_grn,F
	rrf	temp_grn,F	;W = 7/8 grn[n-1] + 1/8 grn[n]

	moviw	FSR1++
	addwf	temp_red,W
	rrf	WREG,W		;W = 1/2 red[n-1] + 1/2 red[n] 
	addwf	temp_red,W
	rrf	WREG,W		;W = 3/4 red[n-1] + 1/4 red[n]
	addwf	temp_red,F
	rrf	temp_red,F	;7/8 red[n-1] + 1/8 red[n]

	moviw	FSR1++
	addwf	temp_blue,W
	rrf	WREG,W		;W = 1/2 blue[n-1] + 1/2 blue[n]
	addwf	temp_blue,W
	rrf	WREG,W		;W = 3/4 blue[n-1] + 1/4 blue[n]
	addwf	temp_blue,F
	rrf	temp_blue,F	;7/8 blue[n-1] + 1/8 blue[n]

	return

;pattern pos. 0: weight is 100% current RGB values
ustep0
	moviw	FSR1++
	movwf	temp_grn
	moviw	FSR1++
	movwf	temp_red
	moviw	FSR1++
	movwf	temp_blue
	return


;pattern pos. +1/8: weight is 7/8 current + 1/8 previous
ustep1	
	moviw	FSR1++
	addwf	temp_grn,F
	rrf	temp_grn,F	;1/2 grn[n-1] + 1/2 grn[n] 
	addwf	temp_grn,F
	rrf	temp_grn,F	;1/4 grn[n-1] + 3/4 grn[n]
	addwf	temp_grn,F
	rrf	temp_grn,F	;1/8 grn[n-1] + 7/8 grn[n]

	moviw	FSR1++
	addwf	temp_red,F
	rrf	temp_red,F	;1/2 red[n-1] + 1/2 red[n] 
	addwf	temp_red,F
	rrf	temp_red,F	;1/4 red[n-1] + 3/4 red[n]
	addwf	temp_red,F
	rrf	temp_red,F	;1/8 red[n-1] + 7/8 red[n]

	moviw	FSR1++
	addwf	temp_blue,F
	rrf	temp_blue,F	;1/2 blue[n-1] + 1/2 blue[n] 
	addwf	temp_blue,F
	rrf	temp_blue,F	;1/4 blue[n-1] + 3/4 blue[n]
	addwf	temp_blue,F
	rrf	temp_blue,F	;1/8 blue[n-1] + 7/8 blue[n]

	return

;pattern pos. +1/4: weight is 3/4 current + 1/4 previous
ustep2
	moviw	FSR1++
	addwf	temp_grn,F
	rrf	temp_grn,F	;1/2 grn[n-1] + 1/2 grn[n] 
	addwf	temp_grn,F
	rrf	temp_grn,F	;1/4 grn[n-1] + 3/4 grn[n]

	moviw	FSR1++
	addwf	temp_red,F
	rrf	temp_red,F	;1/2 red[n-1] + 1/2 red[n] 
	addwf	temp_red,F
	rrf	temp_red,F	;1/4 red[n-1] + 3/4 red[n]

	moviw	FSR1++
	addwf	temp_blue,F
	rrf	temp_blue,F	;1/2 blue[n-1] + 1/2 blue[n] 
	addwf	temp_blue,F
	rrf	temp_blue,F	;1/4 blue[n-1] + 3/4 blue[n]

	return

;pattern pos. +3/8: weight is 5/8 current + 3/8 previous
ustep3
	moviw	FSR1++
	addwf	temp_grn,F
	rrf	temp_grn,F	;1/2 grn[n-1] + 1/2 grn[n] 
	addwf	temp_grn,W
	rrf	WREG,W		;W= 1/4 grn[n-1] + 3/4 grn[n]
	addwf	temp_grn,F
	rrf	temp_grn,F	;= 3/8 grn[n-1] + 5/8 grn[n]

	moviw	FSR1++
	addwf	temp_red,F
	rrf	temp_red,F	;1/2 red[n-1] + 1/2 red[n] 
	addwf	temp_red,W
	rrf	WREG,W		;W = 1/4 red[n-1] + 3/4 red[n]
	addwf	temp_red,F
	rrf	temp_red,F	;3/8 red[n-1] + 5/8 red[n]

	moviw	FSR1++
	addwf	temp_blue,F
	rrf	temp_blue,F	;1/2 blue[n-1] + 1/2 blue[n]
	addwf	temp_blue,W
	rrf	WREG,W		;W = 1/4 blue[n-1] + 3/4 blue[n]
	addwf	temp_blue,F
	rrf	temp_blue,F	;3/8 blue[n-1] + 5/8 blue[n]

	return

;pattern pos. +1/2: weight is 1/2 current + 1/2 previous
ustep4
	moviw	FSR1++
	addwf	temp_grn,F
	rrf	temp_grn,F	;1/2 grn[n-1] + 1/2 grn[n] 

	moviw	FSR1++
	addwf	temp_red,F
	rrf	temp_red,F	;1/2 red[n-1] + 1/2 red[n] 

	moviw	FSR1++
	addwf	temp_blue,F
	rrf	temp_blue,F	;1/2 blue[n-1] + 1/2 blue[n] 

	return


;pattern pos. +5/8: weight is 3/8 current + 5/8 previous
ustep5
	moviw	FSR1++
	addwf	temp_grn,W
	rrf	WREG,W		;W = 1/2 grn[n-1] + 1/2 grn[n] 
	addwf	temp_grn,F
	rrf	temp_grn,F	;3/4 grn[n-1] + 1/4 grn[n]
	addwf	temp_grn,F
	rrf	temp_grn,F	;= 5/8 grn[n-1] + 3/8 grn[n]

	moviw	FSR1++
	addwf	temp_red,W
	rrf	WREG,W		;W = 1/2 red[n-1] + 1/2 red[n] 
	addwf	temp_red,F
	rrf	temp_red,F	;3/4 red[n-1] + 1/4 red[n]
	addwf	temp_red,F
	rrf	temp_red,F	;3/8 red[n-1] + 5/8 red[n]

	moviw	FSR1++
	addwf	temp_blue,W
	rrf	WREG,W		;W = 1/2 blue[n-1] + 1/2 blue[n]
	addwf	temp_blue,F
	rrf	temp_blue,F	;3/4 blue[n-1] + 1/4 blue[n]
	addwf	temp_blue,F
	rrf	temp_blue,F	;5/8 blue[n-1] + 3/8 blue[n]

	return

;pattern pos. +3/4: weight is 1/4 current + 3/4 previous
ustep6
	moviw	FSR1++
	addwf	temp_grn,W
	rrf	WREG,W		;W = 1/2 grn[n-1] + 1/2 grn[n] 
	addwf	temp_grn,F
	rrf	temp_grn,F	;3/4 grn[n-1] + 1/4 grn[n]

	moviw	FSR1++
	addwf	temp_red,W
	rrf	WREG,W		;W = 1/2 red[n-1] + 1/2 red[n] 
	addwf	temp_red,F
	rrf	temp_red,F	;3/4 red[n-1] + 1/4 red[n]

	moviw	FSR1++
	addwf	temp_blue,W
	rrf	WREG,W		;W = 1/2 blue[n-1] + 1/2 blue[n]
	addwf	temp_blue,F
	rrf	temp_blue,F	;3/4 blue[n-1] + 1/4 blue[n]

	return

