;=============================================== 
;8x8 Fast Multiply Subroutine
;Computes pp_<col> = pp_<col> * temp_<col> (col = all 3 colours)
;Numbers are fix-point notation 0.xxxxxxxx
;Based on Formula: a*b = ((a+b)^2)/4 - ((a-b)^2)/4  
;Precision is reduced to increase speed and to reduce size of 
;lookup table to 256 rather than 512 entries.

pp_mulsub
	movlw	high sqtbl
	movwf	FSR1H

;process green:
	movf	pp_grn,W
	addwf	temp_grn,W
	rrf	WREG,W
	movwf	FSR1L
	movf	INDF1,W
	movwf	temp		;= MSB of ((a+b)^2)/4 

	movf	pp_grn,W
	subwf	temp_grn,W	;W = a - b
	btfss	STATUS,C
	sublw	0		;<0: 2th-complement -> abs(va-vb)
	lsrf	WREG,W
	movwf	FSR1L		
	movf	INDF1,W		;= MSB of ((va-vb)^2(/4
	subwf	temp,W 
	movwf	pp_grn

;process red:
	movf	pp_red,W
	addwf	temp_red,W
	rrf	WREG,W
	movwf	FSR1L
	movf	INDF1,W
	movwf	temp		;= MSB of ((a+b)^2)/4 

	movf	pp_red,W
	subwf	temp_red,W	;W = a - b
	btfss	STATUS,C
	sublw	0		;<0: 2th-complement -> abs(va-vb)
	lsrf	WREG,W
	movwf	FSR1L		
	movf	INDF1,W		;= MSB of ((va-vb)^2(/4
	subwf	temp,W 
	movwf	pp_red

;process blue:
	movf	pp_blue,W
	addwf	temp_blue,W
	rrf	WREG,W
	movwf	FSR1L
	movf	INDF1,W
	movwf	temp		;= MSB of ((a+b)^2)/4 

	movf	pp_blue,W
	subwf	temp_blue,W	;W = a - b
	btfss	STATUS,C
	sublw	0		;<0: 2th-complement -> abs(va-vb)
	lsrf	WREG,W
	movwf	FSR1L		
	movf	INDF1,W		;= MSB of ((va-vb)^2(/4
	subwf	temp,W 
	movwf	pp_blue

	return
	

	ORG	0xF00		;last page of program memory used for table
sqtbl
 variable i=0
 while i<256
	retlw high (i * i + 0x80)
i+=1
 endw



